﻿//KML reference
//<styleUrl> 
//URL of a <Style> or <StyleMap> defined in a Document. If the style is in the same file, use a # reference. 
//If the style is defined in an external file, use a full URL along with # referencing. Examples are 
//<styleUrl>#myIconStyleID</styleUrl>
//<styleUrl>http://someserver.com/somestylefile.xml#restaurant</styleUrl>
//<styleUrl>eateries.kml#my-lunch-spot</styleUrl>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;
using System.Xml;
using System.Net;
using System.Web;
using System.IO;
using System.Globalization;
using ThinkGeo.MapSuite.Core;

namespace KmlExtension
{
    public class KmlFeatureSource : FeatureSource
    {
        private const string presetXml = "<?xml version='1.0' encoding='UTF-8'?><kml xmlns='http://earth.google.com/kml/2.1'></kml>";
        private XmlDocument kmlDocument;
        private InMemoryFeatureLayer inMemoryFeatureLayer;
        private Collection<Feature> tempFeatures;
        private KMLFolderData rootFolder = new KMLFolderData();
        private Dictionary<string, Dictionary<string, string>> globalStyles;
        public DateTime Time = new DateTime(2011, 10, 20, 0, 0, 0);
        public DateTime beginTime = new DateTime();
        public DateTime endTime = new DateTime();
        public Boolean process = new Boolean();
        public Collection<string> data = new Collection<string>();
        public string tempBeginTime;
        public string tempEndTime;
        public string fName;
        public int id = 0;
        public int fid = 0;
        public bool addAllFeatures = false;
        //public ValueStyle styles = new ValueStyle();
        public string iconString;
        
        public PointStyle tempPoint = new PointStyle();
        public LineStyle tempLine = new LineStyle();

        public KMLFolderData RootFolder
        {
            get { return rootFolder; }
        }

        public KmlFeatureSource()
        {
        }
        
        /*
        protected KmlFeatureSource()
            : this(string.Empty)
        {

        }

        public KmlFeatureSource(string kmlPathFileName)
            : this(kmlPathFileName, KmlStringType.File)
        {
        }
        /*
        public KmlFeatureSource(Int32 time)
            : this(time)
        {
        }
        */
        public KmlFeatureSource(string kmlString, KmlStringType stringType)
            : base()
        {
            

            switch (stringType)
            {
                case KmlStringType.String:
                    kmlDocument = new XmlDocument();
                    if (string.IsNullOrEmpty(kmlString))
                    {
                        kmlDocument.LoadXml(presetXml);
                    }
                    else
                    {
                        kmlDocument.LoadXml(kmlString);
                    }
                    break;
                case KmlStringType.File:
                default:
                    kmlDocument = new XmlDocument();
                    kmlDocument.Load(kmlString);
                    break;
            }
            this.inMemoryFeatureLayer = new InMemoryFeatureLayer();
            inMemoryFeatureLayer.Open();
            inMemoryFeatureLayer.Columns.Add(new FeatureSourceColumn("linecolor"));
            inMemoryFeatureLayer.Columns.Add(new FeatureSourceColumn("linewidth"));

            inMemoryFeatureLayer.Close();

            globalStyles = new Dictionary<string, Dictionary<string, string>>();
        }

        private void ProcessFolder(XmlNode folder, XmlNamespaceManager manager)
        {
            //ProcessPlacemark(folder, manager);

            
        }

        public void ProcessFolders(XmlNodeList folders, XmlNamespaceManager manager, Collection<KMLFolderData> folderData)
        {
            foreach (XmlNode Folder in folders)
            {
                // Create a KMLFolderData object for this Folder and initialize its values
                KMLFolderData fd = new KMLFolderData();
                // TO DO: initialize values
                fd.beginId = new Dictionary<int, DateTime>();
                fd.endId = new Dictionary<int, DateTime>();

                fd.Name = (Folder.SelectSingleNode("kk:name", manager).InnerText).ToString();
                fd.tag = id;
                id++;
                XmlNode TimeSpan = Folder.SelectSingleNode("kk:TimeSpan", manager);

                if (TimeSpan == null)
                {
                    fd.HasTimeSpan = false;
                }

                else
                {
                    fd.BeginTime = ProcessTimeSpanBegin(TimeSpan, manager);
                    fd.EndTime = ProcessTimeSpanEnd(TimeSpan, manager);
                    //fd.beginId.Add(id, fd.BeginTime);
                    //fd.endId.Add(id, fd.EndTime);
                    //tempBeginTime = (fd.BeginTime).ToString();
                    //tempEndTime = (fd.EndTime).ToString();
                    fd.HasTimeSpan = true;
                }
                ProcessPlacemark(fd, Folder, manager);
                fd.folderFeatures.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(rootFolder.styles);
                fd.folderFeatures.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
                folderData.Add(fd);

                XmlNodeList subfolders = Folder.SelectNodes("kk:Folder", manager);
                if (subfolders != null)
                    ProcessFolders(subfolders, manager, fd.Subfolders);
            }
        }

        protected override void OpenCore()
        {
            if (inMemoryFeatureLayer.InternalFeatures.Count == 0)
            {
                tempFeatures = new Collection<Feature>();
                
                XmlNamespaceManager manager = new XmlNamespaceManager(kmlDocument.NameTable);
                manager.AddNamespace("kk", kmlDocument.DocumentElement.NamespaceURI);

                if (kmlDocument.DocumentElement.HasChildNodes)
                {
                    

                    rootFolder.Name = (kmlDocument.DocumentElement.ChildNodes[0].SelectSingleNode("kk:name", manager).InnerText).ToString();
                    rootFolder.tag = id;
                    id++;

                    XmlNodeList folders = kmlDocument.DocumentElement.ChildNodes[0].SelectNodes("kk:Folder", manager);
                    
                    // TO DO: Initialize rootFolder object with info about the root
                    rootFolder.styles.ColumnName = "name";
                    getAllStyles(@"E:\Ticket and posts\fcstWeatherKML\fcstWeather.kml", rootFolder);

                    ProcessFolders(folders, manager, rootFolder.Subfolders);
                    //foreach (XmlNode Folder in folders)
                    //{
                    //    ProcessFolder(Folder, manager);
                    //}
                    /*
                    XmlNodeList placemarks = kmlDocument.SelectNodes("//kk:Placemark", manager);
                    foreach (XmlNode placemark in placemarks)
                    {
                        ProcessPlacemark(placemark, manager);
                    }
                    */
                    /*
                    XmlNodeList Times = kmlDocument.SelectNodes("//kk:TimeSpan", manager);
                    foreach (XmlNode TimeSpan in Times)
                    {
                        ProcessTimeSpan(TimeSpan, manager);
                    }
                    */
                    // There is no Placemark tag, so it may use NetworkLink
                    /*
                    if (placemarks.Count == 0)
                    {
                        XmlNodeList networkLinks = kmlDocument.SelectNodes("//kk:NetworkLink", manager);
                        foreach (XmlNode networkLink in networkLinks)
                        {
                            ProcessNetworkLink(networkLink, manager);
                        }
                    }
                    */
                    //addFeatures(rootFolder);
                }
            }
        }
        /*
        private void addFeatures(KMLFolderData folder, bool bAddAllFeatures = false)
        {
            bool bTemp = false;
            
            // Tests for time and adds features
            if (folder.HasTimeSpan == true || bAddAllFeatures == true)
            {
                if (folder.BeginTime <= Time && folder.EndTime >= Time || bAddAllFeatures == true)
                {
                    foreach (Feature feature in folder.folderFeatures.InternalFeatures)
                    {
                        inMemoryFeatureLayer.InternalFeatures.Add(feature);
                    }

                    if (folder.Subfolders.Count > 0)
                        bTemp = true;
                }
            }

            foreach (KMLFolderData subfolder in folder.Subfolders)
            {
                addFeatures(subfolder, bTemp);
            }
        }
        */
        private DateTime ProcessTimeSpanBegin(XmlNode TimeSpan, XmlNamespaceManager manager)
        {
            XmlNode beginNode = TimeSpan.SelectSingleNode("kk:begin", manager);

            if (beginNode != null)
            {
                beginTime = DateTime.Parse(beginNode.InnerText);
            }

            return beginTime;
        }

        private DateTime ProcessTimeSpanEnd(XmlNode TimeSpan, XmlNamespaceManager manager)
        {
            XmlNode endNode = TimeSpan.SelectSingleNode("kk:end", manager);

            if (endNode != null)
            {
                endTime = DateTime.Parse(endNode.InnerText);
            }

            return (endTime);
        }
        /*
        private void ProcessNetworkLink(XmlNode networkLink, XmlNamespaceManager manager)
        {
            XmlNode hrefNode = networkLink.SelectSingleNode("kk:Url/kk:href", manager);
            string fileUrl = hrefNode.InnerText;
            WebRequest request = WebRequest.Create(fileUrl);
            request.Timeout = 10000;
            WebResponse response = null;
            try
            {
                response = request.GetResponse();
                string disposition = response.Headers["Content-Disposition"];
                if (!string.IsNullOrEmpty(disposition))
                {
                    string filename = @"C:\temp\" + disposition.Split('=')[1].Trim('\"');
                    WebClient webClient = new WebClient();
                    webClient.DownloadFile(fileUrl, filename);
                    if (filename.EndsWith(".kml", StringComparison.InvariantCultureIgnoreCase))
                    {

                    }
                    else if (filename.EndsWith(".kmz", StringComparison.InvariantCultureIgnoreCase))
                    {
                        filename = new KmlHelper().ProcessUnZip(filename);
                    }
                    else
                    {
                        return;
                    }

                    KmlFeatureSource source = new KmlFeatureSource(filename);
                    source.Open();
                    foreach (Feature item in source.inMemoryFeatureLayer.InternalFeatures)
                    {
                        tempFeatures.Add(item);
                    }
                    source.Close();

                }
            }
            catch
            {

            }
            finally
            {
                if (response != null)
                {
                    response.Close();
                }
            }
        }
        */
        private void ProcessPlacemark(KMLFolderData fd, XmlNode placemark, XmlNamespaceManager manager)
        {
            Collection<Feature> tempCollection = new Collection<Feature>();

            
            XmlNodeList placeMarkXML = placemark.SelectNodes("kk:Placemark", manager);

            /*
                * Point --> Done
                * LineString --> Done
                * LinearRing --> Done
                * Polygon
                * MultiGeometry --> Done
                * Model 
                */

            foreach (XmlNode mark in placeMarkXML)
            {
                XmlNode stylesNode = mark.SelectSingleNode("kk:styleUrl", manager);
                Dictionary<string, string> columnValues = new Dictionary<string, string>();
                /*
                XmlNode descriptionNode = placemark.SelectSingleNode("kk:description", manager);
                if (descriptionNode != null)
                {
                    columnValues.Add("popupHTML", descriptionNode.InnerXml);
                }
                */
                processNodes(mark, manager, tempCollection, columnValues, stylesNode, fd);

                XmlNodeList multigeometry = mark.SelectNodes("kk:MultiGeometry", manager);
                foreach (XmlNode geometry in multigeometry)
                {
                    processNodes(geometry, manager, tempCollection, columnValues, stylesNode, fd);
                }

                XmlNodeList polygon = mark.SelectNodes("kk:Polygon", manager);
                foreach (XmlNode geometry in polygon)
                {
                    //processNodes(geometry, manager, tempCollection, columnValues, stylesNode, fd);
                    XmlNodeList outer = geometry.SelectNodes("kk:outerBoundaryIs", manager);
                    foreach (XmlNode line in outer)
                    {
                        processNodes(line, manager, tempCollection, columnValues, stylesNode, fd);
                    }

                    XmlNodeList inner = geometry.SelectNodes("kk:innerBoundaryIs", manager);
                    foreach (XmlNode line in inner)
                    {
                        processNodes(line, manager, tempCollection, columnValues, stylesNode, fd);
                    }
                }
            }
        }

        void processNodes(XmlNode geometry, XmlNamespaceManager manager, Collection<Feature> tempCollection, Dictionary<string, string> columnValues, XmlNode stylesNode, KMLFolderData fd)
        {
            //XmlNodeList polygonList = geometry.SelectNodes("kk:Polygon", manager);
            //foreach (XmlNode node in polygonList)
            //{

            //}

            XmlNodeList linearRingList = geometry.SelectNodes("kk:LinearRing", manager);
            foreach (XmlNode node in linearRingList)
            {
                string value = node.SelectSingleNode("kk:coordinates", manager).InnerText.Trim();
                //TODO: the document says use " ", not "\r\n"
                string[] values = value.Split(new string[] { " ", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                Collection<Vertex> Vertices = new Collection<Vertex>();
                foreach (string item in values)
                {
                    string[] items = item.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                    Vertex vertex = new Vertex(double.Parse(items[0]), double.Parse(items[1]));
                    Vertices.Add(vertex);
                }
                RingShape ringShape = new RingShape(Vertices);
                if (columnValues.Count == 0 && stylesNode != null)
                {
                    string styleUrl = stylesNode.InnerText;
                    if (styleUrl.StartsWith("#"))
                    {
                        try
                        {
                            string name = styleUrl.TrimStart('#').ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {
                        }
                    }

                    else
                    {
                        try
                        {
                            string name = styleUrl.ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {
                        }
                    }
                }
                Feature temp = new Feature(ringShape, columnValues);
                temp.Id = fid.ToString();
                fid++;
                tempCollection.Add(temp);
            }

            XmlNodeList pointList = geometry.SelectNodes("kk:Point", manager);
            foreach (XmlNode item in pointList)
            {
                string pointValue = item.SelectSingleNode("kk:coordinates", manager).InnerText.Trim();
                string[] pointValues = pointValue.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                PointShape point = new PointShape(double.Parse(pointValues[0]), double.Parse(pointValues[1]));
                if (columnValues.Count == 0 && stylesNode != null)
                {
                    string styleUrl = stylesNode.InnerText;
                    if (styleUrl.StartsWith("#"))
                    {
                        try
                        {
                            string name = styleUrl.TrimStart('#').ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {
                        }
                    }
                }
                Feature temp = new Feature(point, columnValues);
                temp.Id = fid.ToString();
                fid++;
                tempCollection.Add(temp);
            }

            XmlNodeList lineStringList = geometry.SelectNodes("kk:LineString", manager);
            foreach (XmlNode node in lineStringList)
            {
                string value = node.SelectSingleNode("kk:coordinates", manager).InnerText.Trim();
                //TODO: the document says use " ", not "\r\n"
                string[] values = value.Split(new string[] { " ", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                Collection<Vertex> Vertices = new Collection<Vertex>();
                foreach (string item in values)
                {
                    string[] items = item.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                    Vertex vertex = new Vertex(double.Parse(items[0]), double.Parse(items[1]));
                    Vertices.Add(vertex);
                }
                LineShape lineShape = new LineShape(Vertices);

                if (columnValues.Count == 0 && stylesNode != null)
                {
                    string styleUrl = stylesNode.InnerText;
                    if (styleUrl.StartsWith("#"))
                    {
                        try
                        {
                            string name = styleUrl.TrimStart('#').ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {
                                
                        }
                    }
                }
                Feature temp = new Feature(lineShape, columnValues);
                temp.Id = fid.ToString();
                fid++;
                tempCollection.Add(temp);
            }
            foreach (Feature item in tempCollection)
            {
                WellKnownType wellKnownType = item.GetWellKnownType();
                if (wellKnownType == WellKnownType.Polygon || wellKnownType == WellKnownType.Multipolygon ||
                    wellKnownType == WellKnownType.Line || wellKnownType == WellKnownType.Multiline ||
                    wellKnownType == WellKnownType.Point || wellKnownType == WellKnownType.Multipoint)
                {
                    fd.folderFeatures.InternalFeatures.Add(item);
                    inMemoryFeatureLayer.InternalFeatures.Add(item);
                    inMemoryFeatureLayer.FeatureIdsToExclude.Add(item.Id);
                }

            }
            tempCollection.Clear();
        }

        private void getAllStyles(string p, KMLFolderData kmlfd)
        {
            XmlDocument kml = new XmlDocument();
            kml.Load(p);

            XmlNamespaceManager manager = new XmlNamespaceManager(kml.NameTable);
            manager.AddNamespace("kk", kml.DocumentElement.NamespaceURI);

            if (kml.DocumentElement.HasChildNodes)
            {
                XmlNodeList styleList = kml.DocumentElement.ChildNodes[0].SelectNodes("kk:Style", manager);
                foreach (XmlNode item in styleList)
                {
                    /*
                     * <IconStyle>...</IconStyle> --> Done
                       <LabelStyle>...</LabelStyle>
                       <LineStyle>...</LineStyle> --> Done
                       <PolyStyle>...</PolyStyle>
                       <BalloonStyle>...</BalloonStyle> --> N.A.
                       <ListStyle>...</ListStyle>
                     */
                    string styleId = item.Attributes["id"].Value;

                    ValueItem style = new ValueItem();
                    style.Value = styleId;
                    if (item.SelectSingleNode("kk:IconStyle", manager) != null)
                    {
                        XmlNode iconNode = item.SelectSingleNode("kk:IconStyle", manager);
                        XmlNode hrefNode = iconNode.SelectSingleNode("kk:Icon", manager);
                        iconString = hrefNode.SelectSingleNode("kk:href", manager).InnerText.ToString();

                        PointStyle points = new PointStyle();
                        //points.Image = GetImage(iconString);
                        points.PointType = PointType.Bitmap;
                        
                        try
                        {
                            Single heading = Convert.ToSingle(iconNode.SelectSingleNode("kk:heading", manager).InnerText.ToString());
                            points.RotationAngle = heading;
                        }

                        catch (System.NullReferenceException)
                        {
                        }

                        try
                        {
                            double scale = Convert.ToDouble(iconNode.SelectSingleNode("kk:scale", manager).InnerText);
                            points.ImageScale = scale;
                        }

                        catch (System.NullReferenceException)
                        {
                        }
                        style.DefaultPointStyle = points;
                        kmlfd.styles.ValueItems.Add(style);
                    }

                    if (item.SelectSingleNode("kk:LineStyle", manager) != null)
                    {
                        LineStyle lines = new LineStyle();
                        GeoPen pen = new GeoPen();
                        GeoColor color = new GeoColor();
                        Int32 width = new Int32();
                        try
                        {
                            XmlNode colorxml = item.SelectSingleNode("kk:LineStyle", manager);
                            XmlNode colornode = colorxml.SelectSingleNode("kk:color", manager);
                            XmlNode widthnode = colorxml.SelectSingleNode("kk:width", manager);
                            string color2 = colornode.InnerText.ToString();
                            width = Int32.Parse(widthnode.InnerText);
                            color = GeoColor.FromWin32(Int32.Parse(color2, NumberStyles.HexNumber));
                        }

                        catch (System.NullReferenceException)
                        {
                        }

                        pen.Color = color;
                        pen.Width = width;
                        lines.CenterPen = pen;
                        style.DefaultLineStyle = lines;
                        kmlfd.styles.ValueItems.Add(style);
                    }

                    if (item.SelectSingleNode("kk:PolyStyle", manager) != null)
                    {
                        AreaStyle poly = new AreaStyle();
                        GeoSolidBrush brush = new GeoSolidBrush();
                        GeoColor color = new GeoColor();
                        try
                        {
                            XmlNode polyxml = item.SelectSingleNode("kk:PolyStyle", manager);
                            XmlNode colorxml = polyxml.SelectSingleNode("kk:fill", manager);
                            XmlNode colornode = polyxml.SelectSingleNode("kk:color", manager);
                            string color2 = colornode.InnerText.ToString();
                            int nA = int.Parse(color2.Substring(0, 2), NumberStyles.HexNumber);
                            int nR = int.Parse(color2.Substring(2, 2), NumberStyles.HexNumber);
                            int nG = int.Parse(color2.Substring(4, 2), NumberStyles.HexNumber);
                            int nB = int.Parse(color2.Substring(6, 2), NumberStyles.HexNumber);
                            color = GeoColor.FromArgb(nA, nR, nG, nB);
                        }

                        catch (System.NullReferenceException)
                        {
                        }

                        brush.Color = color;
                        style.DefaultAreaStyle.FillSolidBrush = brush;
                        kmlfd.styles.ValueItems.Add(style);
                    }
                }
            }
        }

        private GeoImage GetImage(string sURL)
        {
            MemoryStream ms = new MemoryStream(new WebClient().DownloadData(sURL));
            return new GeoImage(ms);
        }

        private Dictionary<string, string> GetColumnValues(XmlNode stylesNode, XmlNamespaceManager manager)
        {
            Dictionary<string, string> columnValues = new Dictionary<string, string>();
            if (stylesNode != null)
            {
                foreach (XmlNode styleNode in stylesNode.ChildNodes)
                {
                    if (styleNode.Name == "LineStyle")
                    {
                        string linecolor = "ffffffff";
                        if (styleNode.SelectSingleNode("kk:color", manager) != null)
                        {
                            linecolor = styleNode.SelectSingleNode("kk:color", manager).InnerText;
                        }
                        string linewidth = "1";
                        if (styleNode.SelectSingleNode("kk:width", manager) != null)
                        {
                            linewidth = styleNode.SelectSingleNode("kk:width", manager).InnerText;
                        }
                        columnValues.Add("linecolor", linecolor);
                        columnValues.Add("linewidth", linewidth);
                        break;
                    }
                }
            }

            return columnValues;
        }

        protected override void CloseCore()
        {
            //inMemoryFeatureLayer.InternalFeatures.Clear();
            //globalStyles.Clear();
        }

        protected override Collection<Feature> GetAllFeaturesCore(IEnumerable<string> returningColumnNames)
        {
            return inMemoryFeatureLayer.InternalFeatures;
            //return rootFolder.Subfolders[0].Subfolders[0].folderFeatures.InternalFeatures;
        }
            
    }

    public class KMLFolderData
    {
        public string FolderID { get; set; }
        public bool Visible { get; set; }
        public bool HasTimeSpan { get; set; }
        public DateTime BeginTime { get; set; }
        public DateTime EndTime { get; set; }
        public Dictionary<Int32, DateTime> beginId { get; set; }
        public Dictionary<Int32, DateTime> endId { get; set; }
        public string Name { get; set; }
        public string StyleUrl { get; set; }
        public Collection<KMLFolderData> Subfolders { get; set; }
        public InMemoryFeatureLayer folderFeatures { get; set; }
        public int tag { get; set; }
        public ValueStyle styles = new ValueStyle(); 
        

        public KMLFolderData()
        {
            Subfolders = new Collection<KMLFolderData>();
            folderFeatures = new InMemoryFeatureLayer();
        }
    }
    
}
