﻿using System.Collections.ObjectModel;
using System.IO;
using System.Net;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Xml;
using KmlExtension;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.WpfDesktopEdition;

namespace StylesWithInMemoryFeatureLayer
{
    /// <summary>
    /// Interaction logic for TestWindow.xaml
    /// </summary>
    public partial class TestWindow : Window
    {
        public InMemoryFeatureLayer viewLayer = new InMemoryFeatureLayer();
        public ValueStyle styles = new ValueStyle();
        public KmlFeatureLayer layer = new KmlFeatureLayer(@"E:\Ticket and posts\fcstWeatherKML\fcstWeather.kml");
        public bool first = true;
        public string iconString;
        public Collection<int> ids = new Collection<int>();
        public LayerOverlay layerOverlay = new LayerOverlay();
        public FeatureSourceColumn column = new FeatureSourceColumn("name");
        public InMemoryFeatureSource source;

        public TestWindow()
        {
            InitializeComponent();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            Collection<GeoImage> geoimages = getAllStyles(@"E:\Ticket and posts\fcstWeatherKML\fcstWeather.kml");

            wpfMap1.MapUnit = GeographyUnit.DecimalDegree;

            InMemoryFeatureLayer layer = new InMemoryFeatureLayer();
            layer.InternalFeatures.Add(new Feature(0, 0));

            layer.ZoomLevelSet.ZoomLevel01.DefaultPointStyle.Image = geoimages[2];
            layer.ZoomLevelSet.ZoomLevel01.DefaultPointStyle.PointType = PointType.Bitmap;
            layer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            layerOverlay.Layers.Add(layer);
            wpfMap1.Background = new SolidColorBrush(Colors.Transparent);

            wpfMap1.Overlays.Add(layerOverlay);
            wpfMap1.CurrentExtent = new RectangleShape(-180, 90, 180, -90);
            wpfMap1.Refresh();
        }

        private Collection<GeoImage> getAllStyles(string p)
        {
            Collection<GeoImage> geoimages = new Collection<GeoImage>();

            XmlDocument kml = new XmlDocument();
            kml.Load(p);

            XmlNamespaceManager manager = new XmlNamespaceManager(kml.NameTable);
            manager.AddNamespace("kk", kml.DocumentElement.NamespaceURI);

            if (kml.DocumentElement.HasChildNodes)
            {
                XmlNodeList styleList = kml.DocumentElement.ChildNodes[0].SelectNodes("kk:Style", manager);
                int counter = 0;
                foreach (XmlNode item in styleList)
                {
                    string styleId = item.Attributes["id"].Value;

                    ValueItem style = new ValueItem();
                    style.Value = styleId;
                    if (item.SelectSingleNode("kk:IconStyle", manager) != null)
                    {
                        XmlNode iconNode = item.SelectSingleNode("kk:IconStyle", manager);
                        XmlNode hrefNode = iconNode.SelectSingleNode("kk:Icon", manager);
                        iconString = hrefNode.SelectSingleNode("kk:href", manager).InnerText.ToString();

                        GeoImage geoImage = GetImage(iconString);
                        geoimages.Add(geoImage);
                        counter++;
                        if (counter == 5)
                        {
                            break;
                        }
                    }
                }
            }
            return geoimages;
        }

        private GeoImage GetImage(string sURL)
        {
            MemoryStream ms = new MemoryStream(new WebClient().DownloadData(sURL));
            return new GeoImage(ms);
        }
    }
}
